// Helper classes required for this command
const Scene = require('Scene');
const Options = require('Options');
const Camera = require('Camera');
const Instance = require('Instance');
const Group = require('Group');
const Polygon_mesh = require('Polygon_mesh');
const Light = require('Light');
const Lightprofile = require('Lightprofile');
const Image = require('Image');
const Texture = require('Texture');
const Mdl_material_instance = require('Mdl_material_instance');
const Mdl_function_call = require('Mdl_function_call');
const Mdl_attached_function = require('Mdl_attached_function');

// Command definition
module.exports.command = {
    name: 'tutorial_scene_maker',
    description: 'Creates and sets up a scene ready for rendering.',
    groups: ['tutorial', 'javascript'],
    arguments: {
        scene_name: {
            description: 'The name to use when storing the created scene in the database. '
                + 'Other element names will be derived from this name as well.',
            type: 'String'
        },
        export_filename: {
            description: 'If a non-empty string will use this filename to export the created '
                + 'scene at the end of the command.',
            type: 'String',
            default: null
        }
    },
    returns: {
        type: 'Map',
        description: 'Map containing the database names of the options, camera, camera instance and rootgroup.'
    },
    execute: function({scene_name, export_filename}) {
        //
        // Create and empty scene with required elements
        //

        // Create all of the needed database elements
        const scene = new Scene(scene_name, true);
        const options = new Options(`${scene_name}_opt`, true);
        const camera = new Camera(`${scene_name}_cam`, true);
        const camera_instance = new Instance(`${scene_name}_cam_inst`, true);
        const rootgroup = new Group(`${scene_name}_root`, true);

        // Attach the camera to its instance
        camera_instance.item = camera;

        // Attach the camera instance to the rootgroup
        rootgroup.attach(camera_instance);

        // Set default filtering to Gauss with a radius of 1.5 pixels
        options.attributes.set('filter', 2, 'Sint32');
        options.attributes.set('radius', 1.5, 'Float32');

        // Set some rendering termination conditions
        options.attributes.set('progressive_rendering_max_samples', 250, 'Sint32');
        options.attributes.set('progressive_rendering_max_time', Number.MAX_SAFE_INTEGER, 'Sint32');
        options.attributes.set('progressive_rendering_quality_enabled', false, 'Boolean');
        options.attributes.set('progressive_rendering_quality', 1, 'Float32');
        options.attributes.set('progressive_rendering_converged_pixel_ratio', 0.95, 'Float32');

        // Turn on AI denoising
        options.attributes.set('post_denoiser_available', true, 'Boolean');
        options.attributes.set('post_denoiser_enabled', true, 'Boolean');
        options.attributes.set('post_denoiser_start_iteration', 10, 'Sint32');

        // Set some Iray Interactive options in case irt is used
        options.attributes.set('irt_indirect_light_mode', 2, 'Sint32');

        // Set the needed elements onto the scene
        scene.options = options;
        scene.camera_instance = camera_instance;
        scene.root_group = rootgroup;

        //
        // Setup tonemapping on the camera suitable for daylight
        //

        // Select the photographic tonemapping operator
        camera.attributes.set('tm_tonemapper', 'mia_exposure_photographic', 'String');
        camera.attributes.set('tm_enable_tonemapper', true, 'Boolean');

        // Provide all settings for the operator
        camera.attributes.set('mip_cm2_factor', 1.0, 'Float32');
        camera.attributes.set('mip_whitepoint', new RS.Math.Color(1,1,1), 'Color');
        camera.attributes.set('mip_film_iso', 100.0, 'Float32');
        camera.attributes.set('mip_camera_shutter', 250.0, 'Float32');
        camera.attributes.set('mip_f_number', 2.8, 'Float32');
        camera.attributes.set('mip_vignetting', 0.0, 'Float32');
        camera.attributes.set('mip_crush_blacks', 0.2, 'Float32');
        camera.attributes.set('mip_burn_highlights', 0.25, 'Float32');
        camera.attributes.set('mip_burn_highlights_per_component', true, 'Boolean');
        camera.attributes.set('mip_burn_highlights_max_component', false, 'Boolean');
        camera.attributes.set('mip_saturation', 1.0, 'Float32');
        camera.attributes.set('mip_gamma', 2.2, 'Float32');

        //
        // Setup perspective camera parameters and move the camera
        //

        // Set some basic camera parameters for a prespective camera
        camera.resolution_x = 1024;
        camera.resolution_y = 1024;
        camera.aspect = 1.0;
        camera.focal = 35.0;
        camera.aperture = 44.595;

        // Set the camera transformation from a matrix
        camera_instance.matrix = new RS.Math.Matrix4x4({
            xx: 0.0,  xy: 0.0,  xz: 1.0,  xw: 0.0,
            yx: 0.0,  yy: 1.0,  yz: 0.0,  yw: 0.0,
            zx: -1.0, zy: 0.0,  zz: 0.0,  zw: 0.0,
            wx: 1.5,  wy: -1.0, wz: -4.0, ww: 1.0
        });

        //
        // Create daylight environment lighting
        //

        // Import base.mdl to access environment_spherical MDL function
        // NOTE: This file doesn't actually exist, it is built in
        Scene.import_elements('${shader}/base.mdl');

        // Create an MDL function for the physically based daylight system
        let environment = Mdl_function_call.create(`${scene.name}_environment_function`,
            'mdl::base::sun_and_sky', {
                on: true,
                multiplier: 0.10132,
                rgb_unit_conversion: { r: 1.0, g: 1.0, b: 1.0 },
                haze: 0.5,
                redblueshift: 0.0,
                saturation: 1.0,
                horizon_height: 0.001,
                horizon_blur: 0.1,
                ground_color: { r: 0.4, g: 0.4, b: 0.4 },
                night_color: { r: 0.0, g: 0.0, b: 0.0 },
                sun_direction: { x: -0.4545, y: 0.4545, z: -0.0909 },
                sun_disk_intensity: 1.0,
                sun_disk_scale: 1.0,
                sun_glow_intensity: 1.0,
                y_is_up: true,
                flags: 0,
                physically_scaled_sun: true
            }
        );

        // Set the environment_function attribute on the options to actually use the environment
        options.attributes.set('environment_function', environment.name, 'Ref');

        // Set environment intensity to 1.0 as we want to control with the sun sky system
        options.attributes.set('environment_function_intensity', 1.0, 'Float32');

        //
        // Create some artificial light sources in the scene
        //

        // Create a new light element
        let panel_light = new Light('panel_light', true);

        // Set the light type and area source properties
        panel_light.type = 'point';
        panel_light.area_shape = 'rectangle';
        panel_light.area_size = new RS.Math.Vector2(1.2, 0.3);

        // Make the light directly visible
        panel_light.attributes.set('visible', true, 'Boolean');

        // Create a light profile for the light photometry and load the profile
        let light_profile_wallwasher = new Lightprofile('panel_light_profile', true);
        light_profile_wallwasher.filename = 'wallwasher.ies';

        // Load the MDL material needed for photometric lights
        Scene.import_elements('${shader}/material_examples/lights_photometric.mdl');

        // Create the material instance for the photometric light
        let panel_light_mtl = Mdl_material_instance.create('panel_light_mtl', 'mdl::material_examples::lights_photometric::ies_light', {
            flux: 5000,
            profile: light_profile_wallwasher
        });

        // Create an instance of the panel light
        let panel_light_instance = new Instance('panel_light_inst', true);
        panel_light_instance.item = panel_light;

        // Transform the light into position
        let light_rotation_1 = new RS.Math.Matrix4x4();
        let light_rotation_2 = new RS.Math.Matrix4x4();
        let light_translation = new RS.Math.Matrix4x4();
        light_translation.set_translation(-1.5, -2.495, -1.5);
        light_rotation_1.set_rotation(new RS.Math.Vector4(1,0,0,0), 90 * (Math.PI / 180.0));
        light_rotation_2.set_rotation(new RS.Math.Vector4(0,0,1,0), 90 * (Math.PI / 180.0));
        panel_light_instance.matrix = light_translation.clone().multiply(
            light_rotation_1.clone().multiply(light_rotation_2));

        // Assign the light material to the light instance
        panel_light_instance.material = panel_light_mtl;

        // Attach the light to the rootgroup so it renders
        rootgroup.attach(panel_light_instance);

        // Create a new light element
        let disc_light = new Light('disc_light', true);

        // Set the light type and area source properties
        disc_light.type = 'point';
        disc_light.area_shape = 'disc';
        disc_light.area_radius = 0.05;

        // Make the light directly visible
        disc_light.attributes.set('visible', true, 'Boolean');

        // Create a light profile for the light photometry and load the profile
        let light_profile_downlight = new Lightprofile('disc_light_profile', true);
        light_profile_downlight.filename = 'downlight.ies';

        // Create the material instance for the photometric light
        let disc_light_mtl = Mdl_material_instance.create('disc_light_mtl', 'mdl::material_examples::lights_photometric::ies_light', {
            flux: 1200,
            profile: light_profile_downlight
        });

        // Create four disc lights and position them
        let disc_light_instance_1 = new Instance('disc_light_inst_1', true);
        disc_light_instance_1.item = disc_light;
        disc_light_instance_1.material = disc_light_mtl;
        light_translation.set_translation(-0.5, -2.495, -0.5);
        light_rotation_1.set_rotation(new RS.Math.Vector4(1,0,0,0), 90 * (Math.PI / 180.0));
        disc_light_instance_1.matrix = light_translation.clone().multiply(light_rotation_1);
        rootgroup.attach(disc_light_instance_1);

        let disc_light_instance_2 = new Instance('disc_light_inst_2', true);
        disc_light_instance_2.item = disc_light;
        disc_light_instance_2.material = disc_light_mtl;
        light_translation.set_translation(-0.5, -2.495, -2.5);
        light_rotation_1.set_rotation(new RS.Math.Vector4(1,0,0,0), 90 * (Math.PI / 180.0));
        disc_light_instance_2.matrix = light_translation.clone().multiply(light_rotation_1);
        rootgroup.attach(disc_light_instance_2);

        let disc_light_instance_3 = new Instance('disc_light_inst_3', true);
        disc_light_instance_3.item = disc_light;
        disc_light_instance_3.material = disc_light_mtl;
        light_translation.set_translation(-2.0, -2.495, -0.5);
        light_rotation_1.set_rotation(new RS.Math.Vector4(1,0,0,0), 90 * (Math.PI / 180.0));
        disc_light_instance_3.matrix = light_translation.clone().multiply(light_rotation_1);
        rootgroup.attach(disc_light_instance_3);

        let disc_light_instance_4 = new Instance('disc_light_inst_4', true);
        disc_light_instance_4.item = disc_light;
        disc_light_instance_4.material = disc_light_mtl;
        light_translation.set_translation(-2.0, -2.495, -2.5);
        light_rotation_1.set_rotation(new RS.Math.Vector4(1,0,0,0), 90 * (Math.PI / 180.0));
        disc_light_instance_4.matrix = light_translation.clone().multiply(light_rotation_1);
        rootgroup.attach(disc_light_instance_4);

        //
        // Create a 5x3x2 room mesh with a hole representing a window and materials
        //

        // Mesh data needed to generate the room
        let mesh = {
            tagged: true,
            vectors: {
                points: [
                    // Points for the side walls, floor and celing
                    { x: 0.0, y: 0.0, z: 0.0 }, // 0
                    { x: 5.0, y: 0.0, z: 0.0 }, // 1
                    { x: 5.0, y: 0.0, z: 3.0 }, // 2
                    { x: 0.0, y: 0.0, z: 3.0 }, // 3
                    { x: 0.0, y: 2.5, z: 0.0 }, // 4
                    { x: 0.0, y: 2.5, z: 3.0 }, // 5
                    { x: 5.0, y: 2.5, z: 3.0 }, // 6
                    { x: 5.0, y: 2.5, z: 0.0 }, // 7
                    // Points for the hole
                    { x: 0.0, y: 0.5, z: 1.0 }, // 8
                    { x: 0.0, y: 0.5, z: 2.0 }, // 9
                    { x: 0.0, y: 2.0, z: 2.0 }, // 10
                    { x: 0.0, y: 2.0, z: 1.0 }, // 11
                    // For front and back walls
                    { x: 0.0, y: 0.0, z: 0.0 }, // 12
                    { x: 5.0, y: 0.0, z: 0.0 }, // 13
                    { x: 5.0, y: 0.0, z: 3.0 }, // 14
                    { x: 0.0, y: 0.0, z: 3.0 }, // 15
                    { x: 0.0, y: 2.5, z: 0.0 }, // 16
                    { x: 0.0, y: 2.5, z: 3.0 }, // 17
                    { x: 5.0, y: 2.5, z: 3.0 }, // 18
                    { x: 5.0, y: 2.5, z: 0.0 }  // 19
                ],
                normals: [
                    { x: 0.0,  y: 1.0,  z: 0.0}, // 0
                    { x: 0.0,  y: -1.0, z: 0.0}, // 1
                    { x: 1.0,  y: 0.0,  z: 0.0}, // 2
                    { x: -1.0, y: 0.0,  z: 0.0}, // 3
                    { x: 0.0,  y: 0.0,  z: 1.0}, // 4
                    { x: 0.0,  y: 0.0,  z: -1.0} // 5
                ],
                uvs: [
                    { x: 0.0, y: 0.0 }, // 0
                    { x: 5.0, y: 0.0 }, // 1
                    { x: 5.0, y: 3.0 }, // 2
                    { x: 0.0, y: 3.0 }, // 3
                    { x: 3.0, y: 0.0 }, // 4
                    { x: 3.0, y: 2.5 }, // 5
                    { x: 0.0, y: 2.5 }, // 6
                    { x: 5.0, y: 2.5 }, // 7
                    { x: 1.0, y: 0.5 }, // 8
                    { x: 2.0, y: 0.5 }, // 9
                    { x: 2.0, y: 2.0 }, // 10
                    { x: 1.0, y: 2.0 }  // 11
                ]
            },
            vertices: [
                // Floor
                { v: 0, n: 0, t: 0 }, // 0
                { v: 1, n: 0, t: 1 }, // 1
                { v: 2, n: 0, t: 2 }, // 2
                { v: 3, n: 0, t: 3 }, // 3
                // Ceiling
                { v: 4, n: 1, t: 0 }, // 4
                { v: 5, n: 1, t: 3 }, // 5
                { v: 6, n: 1, t: 2 }, // 6
                { v: 7, n: 1, t: 1 }, // 7
                // Left wall
                { v: 0, n: 2, t: 0 }, // 8
                { v: 3, n: 2, t: 4 }, // 9
                { v: 5, n: 2, t: 5 }, // 10
                { v: 4, n: 2, t: 6 }, // 11
                // Right wall
                { v: 2, n: 3, t: 0 }, // 12
                { v: 1, n: 3, t: 4 }, // 13
                { v: 7, n: 3, t: 5 }, // 14
                { v: 6, n: 3, t: 6 }, // 15
                // Front wall
                { v: 12, n: 4, t: 0 }, // 16
                { v: 16, n: 4, t: 6 }, // 17
                { v: 19, n: 4, t: 7 }, // 18
                { v: 13, n: 4, t: 1 }, // 19
                // Back wall
                { v: 15, n: 5, t: 0 }, // 20
                { v: 14, n: 5, t: 1 }, // 21
                { v: 18, n: 5, t: 7 }, // 22
                { v: 17, n: 5, t: 6 }, // 23
                // Hole
                { v: 8,  n: 2, t: 8  }, // 24
                { v: 9,  n: 2, t: 9  }, // 25
                { v: 10, n: 2, t: 10 }, // 26
                { v: 11, n: 2, t: 11 }  // 27
            ],
            // Actual polygons, first number is material id then the polygon
            polygons: [
                [0,   0, 1, 2, 3], // Floor
                [1,   4, 5, 6, 7], // Ceiling
                [2,   8, 9, 10, 11], // Left wall
                [2,   12, 13, 14, 15], // Right wall
                [2,   16, 17, 18, 19], // Front wall
                [2,   20, 21, 22, 23] // Back wall
            ],
            // Add hole in left wall for window, first number is polygon index
            holes: [
                [2,   24, 25, 26, 27]
            ]
        };

        // Create the room mesh with the static create method on Polygon_mesh
        let room_mesh = Polygon_mesh.create('room_obj', mesh, true);

        // Instance the mesh and make it visible since it is not visible by default
        let room_instance = new Instance('room_inst', true);
        room_instance.attributes.set('visible', true, 'Boolean');
        room_instance.item = room_mesh;

        // Import the MDL for the materials we want to use
        Scene.import_elements('${shader}/nvidia/core_definitions.mdl');

        // Create a new material for the walls with some defaults
        let wall_mtl = Mdl_material_instance.create('wall_mtl', 'mdl::nvidia::core_definitions::flex_material', {
            reflectivity: 0.0,
            base_color: new RS.Math.Color(0.65, 0.6, 0.6)
        });

        // Create a new material for the ceiling with some defaults
        let ceiling_mtl = Mdl_material_instance.create('ceiling_mtl', 'mdl::nvidia::core_definitions::flex_material', {
            reflectivity: 0.0,
            base_color: new RS.Math.Color(0.5, 0.5, 0.5)
        });

        // Create UV transformation
        let carpet_uv_transform = new Mdl_attached_function(
            Mdl_function_call.create('carpet_uv_transform', 'mdl::base::rotation_translation_scale', {
                scaling: { x: 5.0, y: 5.0, z: 5.0 }
            })
        );
        let carpet_uv = new Mdl_attached_function(
            Mdl_function_call.create('carpet_uv', 'mdl::base::transform_coordinate', {
                transform: carpet_uv_transform
            })
        );

        // Images to use with textures for our carpet
        let carpet_base_image = new Image('floor_mtl_base_image', true);
        carpet_base_image.filename = 'carpet.png';
        let carpet_normal_image = new Image('floor_mtl_nrm_image', true);
        carpet_normal_image.filename = 'carpet_bump_normal.png';

        // Textures to reference the images for our carpet
        let carpet_base_texture = new Texture('floor_mtl_base_texture', true);
        carpet_base_texture.image = carpet_base_image;
        let carpet_normal_texture = new Texture('floor_mtl_nrm_texture', true);
        carpet_normal_texture.image = carpet_normal_image;

        // Lookup functions to connect to the colour inputs of our material
        carpet_base_lookup = Mdl_function_call.create('floor_mtl_base_lookup', 'mdl::base::file_texture', {
            texture: carpet_base_texture,
            uvw: carpet_uv
        });
        carpet_normal_lookup = Mdl_function_call.create('floor_mtl_nrm_lookup',
            'mdl::base::tangent_space_normal_texture', {
            texture: carpet_normal_texture,
            uvw: carpet_uv,
            factor: 1.0
        });

        // Create a new material for the floor
        let floor_mtl = Mdl_material_instance.create('floor_mtl', 'mdl::nvidia::core_definitions::flex_material', {});

        // Edit the reflectivity parameter and attach textures this could also be done
        // when the material instance is created but shown here as an example
        floor_mtl.arguments.set('reflectivity', 0.0);
        floor_mtl.arguments.attach('base_color', carpet_base_lookup, 'tint');
        floor_mtl.arguments.attach('normal', carpet_normal_lookup);

        // Assign the room materials
        room_instance.material = [
            floor_mtl,
            ceiling_mtl,
            wall_mtl
        ];

        // Finally attach the instance to the rootgroup to add it to the scene for rendering
        rootgroup.attach(room_instance);

        //
        // Import and object from OBJ file and add to the scene
        //

        // Import a simple object and fetch its rootgroup
        let import_result = Scene.import_elements('scenes/Tutorial/chair.obj', {
            import_options: {
                no_lights: true,
                no_camera: true,
                no_options: true,
                no_daylight: true,
                invert_transparency: false
            }
        });
        let chair_group = new Group(import_result.rootgroup);

        // Create an instance of the group containing our new object
        let chair_instance = new Instance(`${chair_group.name}_inst`, true);
        chair_instance.item = chair_group;

        //
        // Import some materials from MDL files and create material instances for the object
        //

        // Import a plastic material from an MDL file then create a material instance from definition
        let plastic_import_result = Scene.import_elements('scenes/Tutorial/mdl/red_plastic.mdl', {
            import_options: { list_elements: true }
        });
        let plastic_elements = plastic_import_result.imported_elements
        let plastic_definition_index = plastic_import_result.imported_elements.length - 2;
        let plastic_definition = plastic_elements[plastic_definition_index];
        let plastic_mtl = Mdl_material_instance.create('red_plastic_mtl', plastic_definition, {});

        // Import a aluminium material from an MDL file then create a material instance from definition
        let aluminium_import_result = Scene.import_elements('scenes/Tutorial/mdl/aluminium.mdl', {
            import_options: { list_elements: true }
        });
        let aluminium_elements = aluminium_import_result.imported_elements
        let aluminium_definition_index = aluminium_import_result.imported_elements.length - 2;
        let aluminium_definition = aluminium_elements[aluminium_definition_index];
        let aluminium_mtl = Mdl_material_instance.create('aluminium_mtl', aluminium_definition, {});

        // Import a walnut material from an MDL file then create a material instance from definition
        let walnut_import_result = Scene.import_elements('scenes/Tutorial/mdl/walnut.mdl', {
            import_options: { list_elements: true }
        });
        let walnut_elements = walnut_import_result.imported_elements
        let walnut_definition_index = walnut_import_result.imported_elements.length - 2;
        let walnut_definition = walnut_elements[walnut_definition_index];
        let walnut_mtl = Mdl_material_instance.create('walnut_mtl', walnut_definition, {});

        //
        // Assign new materials to our imported object
        //

        // Set the material attribute on the imported objects root group
        chair_instance.material = [
            walnut_mtl,
            plastic_mtl,
            aluminium_mtl
        ];

        // Set the material attribute to override any further down in the graph
        chair_instance.material_override = true;

        //
        // Create some copies of the object instance and place in the scene
        //

        // Create four copies of the object instance
        let chair_1 = chair_instance.copy(`${chair_instance.name}_01`);
        let chair_2 = chair_instance.copy(`${chair_instance.name}_02`);
        let chair_3 = chair_instance.copy(`${chair_instance.name}_03`);
        let chair_4 = chair_instance.copy(`${chair_instance.name}_04`);

        // Matrices for our translations and rotations
        let translation = new RS.Math.Matrix4x4();
        let rotation = new RS.Math.Matrix4x4();

        // Compute the matrix for chair 1
        translation.set_translation(-1.0, 0.0, -1.0);
        rotation.set_rotation(new RS.Math.Vector4(0,1,0,0), 135.0 * (Math.PI/180.0));

        // Set the matrix on the instance of chair 1 by composing the rotation
        // and translation and use clone() to ensure the original translation 
        // matrix variable is not modified during the operation
        chair_1.matrix = translation.clone().multiply(rotation);

        // Repeat for chair 2 with new values
        translation.set_translation(-1.0, 0.0, -2.0);
        rotation.set_rotation(new RS.Math.Vector4(0,1,0,0), 45.0 * (Math.PI/180.0));
        chair_2.matrix = translation.clone().multiply(rotation);

        // Repeat for chair 3 with new values
        translation.set_translation(-2.0, 0.0, -1.0);
        rotation.set_rotation(new RS.Math.Vector4(0,1,0,0), -135.0 * (Math.PI/180.0));
        chair_3.matrix = translation.clone().multiply(rotation);

        // Repeat for chair 3 with new values
        translation.set_translation(-2.0, 0.0, -2.0);
        rotation.set_rotation(new RS.Math.Vector4(0,1,0,0), -45.0 * (Math.PI/180.0));
        chair_4.matrix = translation.clone().multiply(rotation);

        // Attach the instances of the new object to the scenes rootgroup
        rootgroup.attach(chair_1);
        rootgroup.attach(chair_2);
        rootgroup.attach(chair_3);
        rootgroup.attach(chair_4);

        // Remove the original instance we made as we now have the copies
        chair_instance.delete();

        //
        // Finish up and return the details of the created scene
        //

        // Export the scene data for review in Iray Viewer
        if (export_filename) {
            scene.export(export_filename);
        }

        // Return the details of our configured scene
        return {
            rootgroup: rootgroup.name,
            options: options.name,
            camera: camera.name,
            camera_instance: camera_instance.name
        };
    }
};
